/**
 * TimeScale.java
 */
package UTEStatsViewer;
import GraphicPlotter.*;
import java.awt.*;
import java.awt.event.*;

/**
 * A GraphicPlotter AxisScale for time in microseconds.
 */

public class TimeScale extends AxisScale {
   private static final int tickSpacingMin = 64;
   private static final int tickWidth = 3;
   private static final int tickHeight = 6;
   private static final int minorTickWidth = 1;
   private static final int minorTickHeight = 3;

   Insets extraInsets = new Insets(0, 0, 0, 0);


   // Override AxisScale methods.

   /** Return insets of graphic elements relative to scale. */
   public Insets getExtraInsets() {
      long min = (long)origin;
      long max = (long)(origin + extent);
      FontMetrics fontMetrics = getFontMetrics(getFont());
      extraInsets.left  = fontMetrics.stringWidth(TimeValue.toString(min))/2;
      extraInsets.right = fontMetrics.stringWidth(TimeValue.toString(max))/2;
      return extraInsets;
   }


   /** Transform a user coordinate to a scale coordinate. */
   public String scaleCoordinateOf(double uc) {
      return TimeValue.toString((long)uc);
   }


   /** Transform a scale coordinate to a user coordinate. */
   public double parseScaleCoordinate(StringBuffer sc) {
      return TimeValue.parseTime(sc);
   }



   // Override Component methods.

   public Dimension getMinimumSize() {
      return getPreferredSize();
   }


   public Dimension getPreferredSize() {
      // Need enough height to show the tick marks, the tick values,
      // and some spacing between them.
      FontMetrics fontMetrics = getFontMetrics(getFont());
      Dimension dim =  new Dimension(1,
            tickHeight + 2*fontMetrics.getLeading() +
            fontMetrics.getMaxAscent() + fontMetrics.getMaxDescent());
      return dim;
   }


   public void paint(Graphics g) {
      Dimension mySize = getSize();
      long min = (long)origin;
      long max = (long)(origin + extent);
      double pitch = ((double)(max - min)) /
            (mySize.width - 1 - extraInsets.left - extraInsets.right);
      if ( (max > min) && (mySize.width > 0) ) {
         // draw scale, else not enough defined to draw it
         Rectangle clipBounds = g.getClipBounds();
         if (clipBounds == null) {
            clipBounds = new Rectangle(0, 0, mySize.width, mySize.height);
         }
//System.out.println(getClass().getName() + ".paint(" + g + ") clipBounds = " + clipBounds);

         // draw horizontal line for the axis
         // not doing our own clipping seems to cause problems
         g.drawLine(clipBounds.x, 0, clipBounds.x+clipBounds.width-1, 0);

         // Calculate first estimate of tick increment.
         long tickDelta = Math.round(tickSpacingMin * pitch);

         // Scale estimated tick increment to a value less than 10.
         long tickFactor = 1;
         while (tickDelta > 100) {
            tickDelta /= 10;
            tickFactor *= 10;
         }

         // Round the tick increment up to 1, 5, 10, 50, or 100.
         int minorTickCnt;
         if (tickDelta > 50) {
            tickDelta = 100;
            minorTickCnt = 2;
         } else if (tickDelta > 10) {
            tickDelta = 50;
            minorTickCnt = 5;
         } else if (tickDelta > 5) {
            tickDelta = 10;
            minorTickCnt = 2;
         } else if (tickDelta > 1) {
            tickDelta = 5;
            minorTickCnt = 5;
         } else {
            tickDelta = 1;
            minorTickCnt = 1;
         }

         // Scale the tick increment back to its original range.
         tickDelta *= tickFactor;
         long minorTickDelta = tickDelta / minorTickCnt;

         // Determine range of visible tick values.
         long tickValue = tickValueBefore(tickDelta,
               clipBounds.x, mySize.width);
         long tickValueHi = tickValueBefore(tickDelta,
               clipBounds.x + clipBounds.width, mySize.width);

         // Determine the y position of the tick values.
         Font font = getFont();
         FontMetrics fontMetrics = getFontMetrics(font);
         int stringYPos = tickHeight + fontMetrics.getLeading() +
               fontMetrics.getMaxAscent();

         // For each tick value to the maximum one that is visible.
         while (tickValue <= tickValueHi) {
            // Determine the x position of the tick and draw it.
            int xPos = (int)Math.round((tickValue - min) / pitch) +
                  extraInsets.left;
            drawMajorTick(g, tickValue, xPos, stringYPos, fontMetrics);

            // Draw the minor ticks to next tick value.
            for (int i = 1; i < minorTickCnt; i++) {
               xPos = extraInsets.left + (int)Math.round(
                     ((tickValue + minorTickDelta * i) - min) / pitch);
               g.fillRect(xPos - (minorTickWidth/2), 1,
                    minorTickWidth, minorTickHeight-1);
            }

            // Increment to next tick value
            if (tickValue > (Long.MAX_VALUE - tickDelta)) {
               break;  // can't increment beyond MAX_VALUE
            } 
            tickValue += tickDelta;
         }
         if (tickValue > tickValueHi) {
            // Draw next major tick and value, some of it may be visible
            // even though the location of the tick mark is beyond the
            // clip width. 
            int xPos = (int)Math.round((tickValue - min) / pitch) +
                   extraInsets.left;
            drawMajorTick(g, tickValue, xPos, stringYPos, fontMetrics);
         }
      }
   }


   /** Draw a major tick and its value. */
   private final void drawMajorTick(Graphics g, long tickValue,
          int xPos, int stringYPos, FontMetrics fontMetrics) {
      g.fillRect(xPos - (tickWidth/2), 1, tickWidth, tickHeight-1);
      String valueString = TimeValue.toString(tickValue);
      g.drawString(valueString,
            xPos - fontMetrics.stringWidth(valueString)/2,
            stringYPos);
   }


   /**
    * Return the greatest tick value that appears at or before the given
    * AWT x position for a scale of the given AWT width for the given
    * tick unit.
    */
   private final long tickValueBefore(long tickUnit, int x, int width) {
      long min = (long)origin;
      long max = (long)(origin + extent);
      long value = ((long)((double)(max - min)/width * x)) + min;
      if (value >= 0)
         value = (value / tickUnit) * tickUnit;
      else
         value = ((value - tickUnit + 1) / tickUnit) * tickUnit;
      return value;
   }
}
