import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/*********Class Graph***********/
/*
  This class defines the basic panel on which the actual graph is painted.
  There are no components contained in the panel. The graph is first painted on
  an off-screen image and then that image is drawn on the panel. There is a
  choice between a line graph and a bar graph.
*/


public class Graph extends JPanel
{
    /*----constant class variables----*/
    
    //----size of x-axis in pixels = graph_size.width
    //----size of y-axis in pixels = graph_size.height 
    public final static Dimension graph_size = new Dimension(500,200);
    
    public final static int
	graph_margin = 20,    //--constant pixel margin between graph and
	                      //--panel edge.
	LINE=1,BAR=2;         //--constants for graph type.

    /*----instance variables----*/

    private double       data[];   
    private double   
	x_range_min,     //--the lower time bound in time units.
	x_range_max,     //--the upper time bound in time units.
	x_bin_time,      //--the time allocated to each bin.
	y_range_min = 0, //--min y-value assumed to be 0.
	y_range_max;     //--max y-value. 
	

    /*
      a bin is a pixel representation of the time interval between each value
      recorded.
    */

    private int      
	x_bin_gap,               //--size of bin in pixels. 
	graph_type = Graph.BAR,  //--default is line graph.  
	x_labl_size = 60,        //--min pixel gap between label on x-axis.
        time_flag = 0,           //-- 0=> time in seconds else time in hours
	y_bin_gap = 40,          //--min pixel gap between labels on the
                                 //--y-axis.
	x_scale_index = 0,
	x_scale_string_width = 6;
    private Dimension display_size;   //--total size of panel.
    private Point     origin;         //--x,y pixel coordinates of origin on
                                      //--on panel.
    private Image     virtual_canvas; //--buffered image in memory.
    private Graphics  brush;          //--graphics context for buffered image.
    private Font      font;           //--font type. 
    private int       markd;          //--flag if point selected on graph. It
                                      //--also contains the x coord for that 
                                      //--point.
    
    /*
      constructor -
      requires data set, the lower time bound and upper time bound
    */
    public Graph(double data[],double x_range_min, double x_range_max)
    {
	int i, max_index, min_index, cal_assist1, cal_assist2;
	String temp_string;
	markd = 0;
	font = new Font("Dialog",Font.PLAIN,12);
	this.data = data;
	
	this.x_range_max = x_range_max;
	this.x_range_min = x_range_min;
	if(x_range_max > 3600)
	    {
		time_flag = 1;
		this.x_range_min=this.x_range_min/3600;
		this.x_range_max=this.x_range_max/3600;
	    }

	//** the following paragraph of code generates the scale of the graph
	//** in terms of an integer index of 10 
	//** for the case where max is either too large or too small.

	max_index = (int)(Math.log(this.x_range_max)/Math.log(10.0));
	if(max_index > (x_scale_string_width-1))
	    x_scale_index = max_index - x_scale_string_width;
	else if( max_index < ((-1)*(x_scale_string_width-2)))
	    x_scale_index = (x_scale_string_width-2)+max_index;


	
	this.x_bin_gap = graph_size.width / (data.length);
	if(this.x_bin_gap == 0)
	    {
		System.out.println("Number of values to be graphed exceeds" + 
				   " the maximum display size of 500 pixels");
		System.exit(1);
	    }
	  
	x_bin_time = ((double)x_bin_gap / ((data.length) * x_bin_gap)) *
	    (this.x_range_max - this.x_range_min);

	y_range_min = 0;
	y_range_max = 0;


	//** finding the largest value in the data set to initialize the max
	//** y value. 

	for (i=0;i<data.length;i++)
	    if(data[i] > y_range_max)
		y_range_max = (double)data[i];
	
	//** the max y value is always a bit higher than the actual max data 
	//** value by 20 pixels.
	
	y_range_max += (((double)20 / graph_size.height) * 
			(y_range_max-y_range_min));


	//** making sure that the labels on the y-axis get positioned
	//** right if the highest value is less than the number of pixels
	//** allocated to the graph's height.

	if(y_range_max < graph_size.height)
	    {
		cal_assist1 = (int)(graph_size.height / y_range_max);
		cal_assist2 = (y_bin_gap/cal_assist1) * cal_assist1;
		if(cal_assist2 < y_bin_gap)
		    y_bin_gap = cal_assist2 + cal_assist1;
	    }

	temp_string = Double.toString(y_range_max);
	temp_string = temp_string.substring(0,
					    temp_string.indexOf((int)'.') +
					    2);

	origin = new Point(graph_margin + 6 +
			   getFontMetrics(font).stringWidth(temp_string),
			   graph_margin + graph_size.height);

	display_size = new Dimension(origin.x + 
				     graph_size.width +
				     graph_margin,
				     origin.y + (3 * graph_margin));
	
	setPreferredSize(display_size);
	setMinimumSize(display_size);
	setMaximumSize(display_size);
	setBackground(Color.blue);
    }

	
    /*
      this is where all the drawing takes place as well as the initialization 
      for the graphics context.
    */
    public void paint(Graphics g)
    {
	super.paint(g);
	
	g.setColor(Color.white);

	virtual_canvas = createImage(display_size.width,display_size.height);
	brush = virtual_canvas.getGraphics();

	//** standardizing the font for the axis labels.

	font = brush.getFont();

	//** initializing the label size so as to be a multiple of the bin size
	//** and also to make sure that the labels dont overlap.

	x_labl_size = getFontMetrics(font).stringWidth("digitsize");
	if(x_bin_gap >= x_labl_size)
	    x_labl_size = x_bin_gap;
	else
	    x_labl_size = ((x_labl_size / x_bin_gap) + 1) * x_bin_gap;

	drawAxis();
	
	int i;
	if(graph_type == Graph.BAR)
	    drawBarGraph();
	else
	    drawLineGraph();

	g.drawImage(virtual_canvas,0,0,Color.white,this);

	//** draw the marker if the point was selected

	if(markd != 0)
	    {
		g.setColor(Color.red);
		g.drawLine(markd,origin.y,markd,
			   origin.y-graph_size.height);
		g.setColor(Color.white);
	    }
    }

    /*
      draws both the x-axis and the y-axis with all the labels.
    */
    private void drawAxis()
    {
	brush.setColor(Color.white);
	brush.fillRect
	    (0,0,virtual_canvas.getWidth(this),virtual_canvas.getHeight(this));
	brush.setColor(Color.black);

	//** drawing the axes

	brush.drawLine(origin.x, origin.y, origin.x, graph_margin-5);
	brush.drawLine(origin.x, origin.y, origin.x + graph_size.width+15,
		       origin.y);
	
	//** drawing the arrows

	brush.drawLine(origin.x, graph_margin - 5, origin.x - 5, 
		       graph_margin);
	brush.drawLine(origin.x, graph_margin - 5, origin.x + 5, 
		       graph_margin);
	brush.drawLine(origin.x + graph_size.width + 15, origin.y,
		       origin.x + graph_size.width+10, origin.y - 5);
	brush.drawLine(origin.x + graph_size.width + 15, origin.y,
		       origin.x + graph_size.width+10, origin.y + 5);
	
	int i;

	//** draw labels on x-axis

	for(i=origin.x + x_labl_size;
	    (((i-origin.x)/x_bin_gap)*x_bin_time+x_range_min) < x_range_max; 
	    i+=x_labl_size)
	    {
		brush.drawLine(i,origin.y,i,origin.y+4);
		centerLabelOnX(i);
	    }

	//** draw labels on y-axis

	for(i=origin.y - y_bin_gap; i>(origin.y-(graph_size.height-4));
	    i-=y_bin_gap)
	    {
		brush.drawLine(origin.x,i,origin.x-4,i);
		centerLabelOnY(i);
	    }

	//** draw origin values

	String temp = calculateXValueString(origin.x,3);
	temp = "("+temp+","+y_range_min+")";
	brush.drawString(temp,
                         origin.x - getFontMetrics(font).stringWidth(temp)/2,
			 origin.y + getFontMetrics(font).getMaxAscent() + 6);

	temp = "time ( ";
	if(x_scale_index != 0)
	    temp += "x10    ";
	
	if(time_flag == 0)
	    temp += "seconds )";
	else
	    temp += "hours )";
	

  
	if((x_scale_index !=  0)&&(x_scale_index != 1))
	    {
		brush.setFont(new Font(brush.getFont().getName(),
				       brush.getFont().getStyle(),
				       9));
		brush.drawString(Integer.toString(x_scale_index),
				 origin.x + (graph_size.width/2) - 
				 (getFontMetrics(font).stringWidth(temp)/2)+
				 (getFontMetrics(font).
				  stringWidth("time ( x10")),
				 origin.y +
				 4*getFontMetrics(font).getMaxAscent()-4);
		brush.setFont(font);
	    }

	brush.drawString(temp,
			 origin.x + (graph_size.width/2) - 
			 (getFontMetrics(font).stringWidth(temp)/2),
			 origin.y + 4*getFontMetrics(font).getMaxAscent());

	//** draw the marker to show the max x-value
	brush.setColor(Color.red);
	brush.drawLine(origin.x + (x_bin_gap * data.length),origin.y,
		       origin.x + (x_bin_gap * data.length),origin.y+8);

	temp = calculateXValueString((x_bin_gap*data.length)+origin.x,6);

	brush.drawString(temp,
			 origin.x + (x_bin_gap * data.length)- (int)
			 (0.75*getFontMetrics(font).stringWidth(temp)),
			 origin.y + 2*getFontMetrics(font).getMaxAscent()+8);

	brush.setColor(Color.black);
    }

    /*
      this returns the dimensions of the panel
    */
    public Dimension getDimension()
    {
	return (new Dimension(display_size));
    }

    /*
      centers the label at the giveen position under the x-axis
    */
    private void centerLabelOnX(int x_position)
    {
	String value_s =  calculateXValueString(x_position,6);
	/*double value;
	value = (((double)(x_position-origin.x)/x_bin_gap)*x_bin_time)+
	    x_range_min;
	Double.toString(value);
	if(value_s.length()>6)
	    value_s = value_s.substring(0,5);	
	*/
	brush.drawString(value_s,
		      x_position-(getFontMetrics(font).stringWidth(value_s)/2),
		      (origin.y+6)+(getFontMetrics(font).getMaxAscent()));
    }

    /*
      centers the label at the given position by the y-axis
    */
    private void centerLabelOnY(int y_position)
    {
	double value;
	value = ((((double)(origin.y-y_position)/graph_size.height)*
	    (y_range_max-y_range_min))+(double)y_range_min);
	String value_s = Double.toString(value);
	value_s = value_s.substring(0, value_s.indexOf((int)'.') + 2);
	brush.drawString(value_s,
		      (origin.x-6)-(getFontMetrics(font).stringWidth(value_s)),
		      y_position);
    }
    
    /*
      draws the entire bar graph in one shot
    */
    private void drawBarGraph()
    {
	int i,j,y_coord1=0,bar_width=1;
	brush.setColor(Color.blue);
	
	//** calculating bar width so as not to intersect any two bars 
	if(x_bin_gap > 3)
	    if(x_bin_gap%2 == 0)
		bar_width = (x_bin_gap-1);
	    else bar_width = (x_bin_gap-2);
	/*if(bar_width > 30)
	  bar_width = 30;*/

	for(i=0,j=origin.x+x_bin_gap;
	    i<data.length;i++,j+=x_bin_gap)
	    {
		y_coord1 = origin.y - (int)
		    ((data[i]/y_range_max)*graph_size.height);
		brush.fillRect(j-bar_width+1, y_coord1,
			       bar_width,origin.y-y_coord1);
	    }
	brush.setColor(Color.black);
    }

    /*
      draws line graph in one shot
    */
    private void drawLineGraph()
    {
	int i,j,y_coord1=0,y_coord2;
	brush.setColor(Color.blue);
	
	//** making sure there is data
	if(data.length > 0)
	    y_coord1 = origin.y;

	for(i=0,j=origin.x;i<data.length;i++,j+=x_bin_gap)
	    {
		y_coord2 = origin.y - (int)
		    ((data[i] / (y_range_max-y_range_min))
		     * graph_size.height);
	
		brush.drawLine(j,y_coord1,j+x_bin_gap,y_coord2);

		y_coord1 = y_coord2;
	    }
	brush.setColor(Color.black);
    }

    public double getXBinTime()
    {
	return x_bin_time;
    }

    public int getXBinGap()
    {
	return x_bin_gap;
    }

    public Point getOrigin()
    {
	return new Point(origin.x,origin.y);
    }

    public double getMinTimeVal()
    {
	return x_range_min;
    }
    
    /*
      marks the point selected in the graph by setting the value of 
      marker to the time value corresponding to that point.
    */
    public double mark_time(Point marker)
    {
	markd = marker.x;
	repaint();
	if(x_range_max > 3600)
	    return ((((double)(marker.x-origin.x)/x_bin_gap)*x_bin_time)+
		    x_range_min)*3600;
	else
	    return (((double)(marker.x-origin.x)/x_bin_gap)*x_bin_time)+
		x_range_min;	
    }

    /*
      changes the graph representation
    */
    public void changeGraphType(int type)
    {
	this.graph_type = type;
	repaint();
    }
    
    /*
      returns the range of the graph in pixels from x_min to x_max
    */
    public int getGraphRange()
    {
	return x_bin_gap*data.length;
    }

    /*
      This function returns a string representing the scaled value.
      -precision is a positive number defining the max number of characters
      in the returned string.
      -x_pos is the position on the graph for which a string representation
      of the time is required.
    */
    public String calculateXValueString(int x_pos,int precision)
    {
	String value_s,temp_string;
	double value;
	int indexValue,
	    min_displyd_char = 2;//-- example: "0.0.......1"
	                         //-- the "0." is 2 char long.

	value = (((double)(x_pos-origin.x)/x_bin_gap)*x_bin_time)+
	    x_range_min;

	//**scaling the value obtained from the above operation
	value /= Math.pow(10,x_scale_index);
	
	temp_string = Double.toString(value);

	//** if the value is very small or very large the string representation
	//** is in scientific notation which poses a tricky problem when
	//** the string needs to be displayed on the graph. the solution is 
	//** taken care of in the following if statement.

	if(temp_string.indexOf((int)'E') != -1)
	    {
		indexValue = Integer.parseInt(temp_string.substring
				    (temp_string.indexOf((int)'E')+1));

		//** since the max value has already been scaled down,
		//** only negative indices are possible in the string 
		//** representation of the time value. The following if 
		//** statement checks to make sure if the exponent is within
		//** the prescribed length of the final string (precision).
		if(indexValue >= ((-1)*(precision-min_displyd_char)))
		    //** 
		    {
			value_s ="0.";
			int zero_counter = (-1)*indexValue;
			while(zero_counter > 1)
			    {
				value_s += "0";
				zero_counter--;
			    }
			int i=0;
			value_s += temp_string.charAt(i);
			i=2;
			while((value_s.length() < precision)&&
			      (temp_string.charAt(i) != 'E'))
			    {
				value_s += temp_string.charAt(i);
				i++;
			    }
		    }
		else value_s = "0.0";
	    }
	else
	    {
		value_s = temp_string;
		if(value_s.length()>precision)
		    value_s = value_s.substring(0,precision);
		if(value_s.indexOf((int)'.') == (value_s.length()-1))
		    value_s = value_s.substring(0,value_s.length());
	    }
	return new String(value_s);
    }	
    
    /*
      This function accepts a time value and moves the marker on the graph
      to the cooresponding value on the graph.
    */
    void moveMarkerTo(double time)
    {
	this.markd = (int)((time/this.x_bin_time)*x_bin_gap)+origin.x;
	repaint();
    }
    
}
	
